/**
 * @file stack.c
 * @brief stack tracing
 * @author Ben Blum
 *
 * Copyright (c) 2018, Ben Blum
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define MODULE_NAME "STACK"
#define MODULE_COLOUR COLOUR_DARK COLOUR_BLUE

#include "common.h"
#include "html.h"
#include "kernel_specifics.h"
#include "kspec.h"
#include "landslide.h"
#include "stack.h"
#include "symtable.h"
#include "variable_queue.h"
#include "x86.h"

#define FRAME_LIST_INITIAL_SIZE 16

/******************************************************************************
 * printing utilities / glue
 ******************************************************************************/

#define FRAME_BUF_LEN 256

/* guaranteed not to clobber nobe. */
bool eip_to_frame(unsigned int eip, struct stack_frame *f)
{
	f->eip = eip;
	f->actual_eip = eip;
	f->name = NULL;
	f->file = NULL;
	return symtable_lookup(eip, &f->name, &f->file, &f->line);
}

void destroy_frame(struct stack_frame *f)
{
	if (f->name != NULL) MM_FREE(f->name);
	if (f->file != NULL) MM_FREE(f->file);
}

/* Emits a "0xADDR in NAME (FILE:LINE)" line with optional pretty colours. */
unsigned int sprint_frame(char *buf, unsigned int maxlen,
			  const struct stack_frame *f, bool colours)
{
#define PRINT(...) do { pos += scnprintf(buf + pos, maxlen - pos, __VA_ARGS__); } while (0)
	unsigned int pos = 0;
	PRINT("0x%.8x in ", f->eip);
	if (f->name == NULL) {
		if (colours) {
			PRINT(COLOUR_BOLD COLOUR_MAGENTA);
		}
		PRINT("<");
		if (f->eip < PAGE_SIZE) {
			PRINT("zero-town");
		} else if (USER_MEMORY(f->eip)) {
			PRINT("unknown in userspace");
		} else if (f->eip > GUEST_DATA_START) {
			PRINT("unknown in kernel");
		} else {
			PRINT("almost certainly bogus");
		}
		PRINT(">");
		if (colours) {
			PRINT(COLOUR_DEFAULT);
		}
	} else {
		if (colours) {
			PRINT(COLOUR_BOLD COLOUR_CYAN);
		}
		PRINT("%s ", f->name);
		if (colours) {
			PRINT(COLOUR_DARK COLOUR_GREY);
		}
		if (f->file == NULL) {
			PRINT("<unknown>");
		} else {
			PRINT("(%s:%d)", f->file, f->line);
		}
		if (colours) {
			PRINT(COLOUR_DEFAULT);
		}
	}
	return pos;
#undef PRINT
}

void print_stack_frame(verbosity v, const struct stack_frame *f)
{
	char buf[FRAME_BUF_LEN];
	sprint_frame(buf, FRAME_BUF_LEN, f, true);
	printf(v, "%s", buf);
}

/* Same as print_stack_frame but includes the symbol table lookup. */
void print_eip(verbosity v, unsigned int eip)
{
	struct stack_frame f;
	eip_to_frame(eip, &f);
	print_stack_frame(v, &f);
	destroy_frame(&f);
}

/* Prints a stack trace to the console. Uses printf, not lsprintf, separates
 * frames with ", ", and does not emit a newline at the end. */
void print_stack_trace(verbosity v, const struct stack_trace *st)
{
	const struct stack_frame *f;
	unsigned int i;
	bool first_frame = true;

	/* print TID prefix before first frame */
	printf(v, "TID%d at ", st->tid);

	/* print each frame */
	ARRAY_LIST_FOREACH(&st->frames, i, f) {
		if (!first_frame) {
			printf(v, ", ");
		}
		first_frame = false;
		print_stack_frame(v, f);
	}
}

/* Prints a stack trace to a multiline html table. Returns length printed. */
unsigned int html_stack_trace(char *buf, unsigned int maxlen, const struct stack_trace *st)
{
#define PRINT(...) do { pos += scnprintf(buf + pos, maxlen - pos, __VA_ARGS__); } while (0)
	unsigned int pos = 0;
	bool first_frame = true;
	const struct stack_frame *f;
	unsigned int i;

	ARRAY_LIST_FOREACH(&st->frames, i, f) {
		if (!first_frame) {
			PRINT("<br />");
		}
		first_frame = false;
		/* see print_stack_frame, above */
		PRINT("0x%.8x in ", f->eip);
		if (f->name == NULL) {
			PRINT(HTML_COLOUR_START(HTML_COLOUR_MAGENTA)
			      "&lt;%s&gt;" HTML_COLOUR_END,
			      f->eip < PAGE_SIZE ? "zero-town"
			      : KERNEL_MEMORY(f->eip) ? "unknown in kernel"
			      : "unknown in userspace");
		} else {
			PRINT(HTML_COLOUR_START(HTML_COLOUR_CYAN) "<b>%s</b>"
			      HTML_COLOUR_END " "
			      HTML_COLOUR_START(HTML_COLOUR_GREY) "<small>",
			      f->name);
			if (f->file == NULL) {
				PRINT("&lt;unknown&gt;");
			} else {
				PRINT("(%s:%d)", f->file, f->line);
			}
			PRINT("</small>" HTML_COLOUR_END);
		}
		PRINT("\n");
	}

	return pos;
#undef PRINT
}

struct stack_trace *copy_stack_trace(const struct stack_trace *src)
{
	struct stack_trace *dest = MM_XMALLOC(1, struct stack_trace);
	const struct stack_frame *f_src;
	unsigned int i;
	dest->tid = src->tid;
	ARRAY_LIST_INIT(&dest->frames, FRAME_LIST_INITIAL_SIZE);

	ARRAY_LIST_FOREACH(&src->frames, i, f_src) {
		struct stack_frame f_dest;
		f_dest.eip  = f_src->eip;
		f_dest.actual_eip  = f_src->actual_eip;
		f_dest.name = f_src->name == NULL ? NULL : MM_XSTRDUP(f_src->name);
		f_dest.file = f_src->file == NULL ? NULL : MM_XSTRDUP(f_src->file);
		f_dest.line = f_src->line;
		ARRAY_LIST_APPEND(mutable_stack_frames(dest), f_dest);
	}
	return dest;
}

void free_stack_trace(struct stack_trace *st)
{
	struct stack_frame *f;
	unsigned int i;
	ARRAY_LIST_FOREACH(mutable_stack_frames(st), i, f) {
		destroy_frame(f);
	}
	ARRAY_LIST_FREE(mutable_stack_frames(st));
	MM_FREE(st);
}

static bool splice_pre_vanish_trace(struct ls_state *ls, struct stack_trace *st,
				    unsigned int eip)
{
	struct stack_trace *pvt = mutable_pre_vanish_trace(ls->sched.cur_agent);
	bool found_eip = false;

	if (pvt == NULL || KERNEL_MEMORY(eip)) {
		return false;
	}

	struct stack_frame *f;
	unsigned int i;
	ARRAY_LIST_FOREACH(mutable_stack_frames(pvt), i, f) {
		if (f->eip == eip || f->actual_eip == eip) {
			found_eip = true;
		}
		if (found_eip) {
			struct stack_frame newf;
			newf.eip  = f->eip;
			newf.name = f->name == NULL ? NULL : MM_XSTRDUP(f->name);
			newf.file = f->file == NULL ? NULL : MM_XSTRDUP(f->file);
			newf.line = f->line;
			ARRAY_LIST_APPEND(mutable_stack_frames(st), newf);
		}
	}
	return found_eip;
}

/* Usually stack trace eips are the pushed return address; i.e., the instruction
 * after the "call". But when a 'noreturn' function's last instruction is a call
 * this will result in the subsequent function being shown in the trace, which
 * is of course very confusing. So we special case here... */
static unsigned int check_noreturn_function(cpu_t *cpu, unsigned int eip)
{
	unsigned int eip_offset;
	// FIXME: this should really be a more unified approach in which
	// landslide *always* finds the actual call instruction, or whatever
	// it was; this is hard to do in general though because of page faults
	if (function_eip_offset(eip, &eip_offset) && eip_offset == 0) {
		if (READ_BYTE(cpu, eip - 5) == OPCODE_CALL) {
			return eip - 5;
		} else if (READ_BYTE(cpu, eip - 2) == OPCODE_INT &&
			   (READ_BYTE(cpu, eip - 1) == VANISH_INT ||
			    READ_BYTE(cpu, eip - 1) == TASK_VANISH_INT)) {
			return eip - 2;
		}
	}
	return eip;
}

/******************************************************************************
 * actual logic
 ******************************************************************************/

/* returns false if symtable lookup failed */
static bool add_frame(struct stack_trace *st, unsigned int eip,
		      unsigned int actual_eip)
{
	struct stack_frame f;
	bool lookup_success = eip_to_frame(eip, &f);
	/* also remember what it was before correcting for noreturn callsites */
	f.actual_eip = actual_eip;
	ARRAY_LIST_APPEND(mutable_stack_frames(st), f);
	return lookup_success;
}

/* Suppress stack frames from userspace, if testing userland, unless the
 * verbosity setting is high enough. */
#define SUPPRESS_FRAME(eip) \
	(MAX_VERBOSITY < DEV && testing_userspace() && KERNEL_MEMORY(eip) )

#ifdef PINTOS_KERNEL
/* The pintos boot sequence puts the initial kstack below the kernel image
 * (approx 0xc000eXXX, compared to 0xc002XXXX). */
#define CHECK_JUNK_EBP_BELOW_TEXT(ebp) false
#else
#define CHECK_JUNK_EBP_BELOW_TEXT(ebp) ((unsigned)(ebp) < GUEST_DATA_START)
#endif

struct stack_trace *stack_trace(struct ls_state *ls)
{
	cpu_t *cpu = ls->cpu0;
	unsigned int eip = ls->eip;
	unsigned int tid = ls->sched.cur_agent->tid;

	unsigned int stack_ptr = GET_CPU_ATTR(cpu, esp);

	struct stack_trace *st = MM_XMALLOC(1, struct stack_trace);
	st->tid = tid;
	ARRAY_LIST_INIT(&st->frames, FRAME_LIST_INITIAL_SIZE);

	/* Add current frame, even if it's in kernel and we're in user. */
	add_frame(st, eip, eip);

	unsigned int stop_ebp = 0;
	unsigned int ebp = GET_CPU_ATTR(cpu, ebp);
	unsigned int rabbit = ebp;
	unsigned int frame_count = 0;
	/* used for remembering pushed return addrs before correcting for
	 * noreturn callsites, which may rely in user memory being present */
	unsigned int actual_eip;

	/* Figure out if the thread is vanishing and we should expect its cr3
	 * to have been freed (so we can't trace in userspace).
	 * (Note this condition will also hit init/shell, but we should not
	 * expect unknown symtable results from them in any case.) */
	bool wrong_cr3 =
		testing_userspace() && GET_CR3(cpu) != ls->user_mem.cr3;

	while (ebp != stop_ebp && (testing_userspace() || KERNEL_MEMORY(ebp))
	       && frame_count++ < 1024) {
		bool extra_frame;

		/* This is the same check as at the end (after ebp advances),
		 * but duplicated here for the corner case where ebp's initial
		 * value is trash. (It's inside the loop so we can still look
		 * for extra frames.) */
		if (CHECK_JUNK_EBP_BELOW_TEXT(ebp)) {
			ebp = 0;
		}

		/* Find "extra frames" of functions that don't set up an ebp
		 * frame or of untimely interrupts, before following ebp. */
		do {
			unsigned int eip_offset;
			bool iret_block = false;

			extra_frame = false;
			/* at the beginning or end of a function, there is no
			 * frame, but a return address is still on the stack. */
			if (function_eip_offset(eip, &eip_offset)) {
				if (eip_offset == 0) {
					if (kern_page_fault_handler_entering(eip)) {
						/* Special case because an error
						 * code will appear on the stack
						 * in front of our eip. */
						stack_ptr += WORD_SIZE;
						iret_block = true;
					}
					extra_frame = true;
				} else if (eip_offset == 1 &&
				           READ_BYTE(cpu, eip - 1)
				           == OPCODE_PUSH_EBP) {
					stack_ptr += WORD_SIZE;
					extra_frame = true;
				} else if (eip_offset % 4 == 0) {
					/* Look for assembly such as,
					 * foo: mov 0x4(%esp),bar
					 *      mov 0x8(%esp),baz
					 *      swizzle bar, baz # you are here
					 *      ...
					 * to find the caller. movs such as this
					 * are always 4 bytes, "8B XX XX XX". */
					bool all_movs = true;
					while (eip_offset > 0) {
						if (READ_BYTE(cpu, eip - eip_offset)
						    != OPCODE_MOV_R32_M32) {
							all_movs = false;
							break;
						}
						eip_offset -= 4;
					}
					if (all_movs) {
						extra_frame = true;
					}
				}
			}
			if (!extra_frame) {
				/* Attempt to understand the tail end of syscall
				 * or interrupt wrappers. Traverse pushed GPRs
				 * if necessary to find the ret or iret. */
				unsigned int opcode;
				unsigned int opcode_offset = 0;
				do {
					opcode = READ_BYTE(cpu, eip + opcode_offset);
					opcode_offset++;
					if (opcode == OPCODE_RET) {
						extra_frame = true;
					} else if (opcode == OPCODE_IRET) {
						iret_block = true;
						extra_frame = true;
					} else if (OPCODE_IS_POP_GPR(opcode)) {
						stack_ptr += WORD_SIZE;
					} else if (opcode == OPCODE_POPA) {
						stack_ptr += WORD_SIZE * POPA_WORDS;
					}
				} while (OPCODE_IS_POP_GPR(opcode) ||
					 opcode == OPCODE_POPA);
			}
			if (extra_frame) {
				actual_eip = eip = READ_MEMORY(cpu, stack_ptr);
				eip = check_noreturn_function(cpu, eip);
				if (splice_pre_vanish_trace(ls, st, eip)) {
					return st;
				} else if (!SUPPRESS_FRAME(eip)) {
					bool success = add_frame(st, eip, actual_eip);
					if (!success && wrong_cr3)
						return st;
				}

				/* Keep walking looking for more extra frames. */
				if (!iret_block) {
					/* Normal function call. */
					stack_ptr += WORD_SIZE;
				} else if (READ_MEMORY(cpu, stack_ptr + WORD_SIZE) ==
					   SEGSEL_KERNEL_CS) {
					/* Kernel-to-kernel iret. Look past it. */
					stack_ptr += WORD_SIZE * IRET_BLOCK_WORDS;
				} else if (READ_MEMORY(cpu, stack_ptr + WORD_SIZE) == SEGSEL_USER_CS) {
					/* User-to-kernel iret. Stack switch. */
					unsigned int esp_addr =
						stack_ptr + (3 * WORD_SIZE);
					stack_ptr = READ_MEMORY(cpu, esp_addr);
				} else {
					/* Unknown case (used to be assert fail) */
					lsprintf(DEV, COLOUR_BOLD COLOUR_YELLOW
						 "Warning: unhandled case in "
						 "stack trace; truncating.\n");
					return st;
				}
			}
		} while (extra_frame);

		if (ebp == 0) {
			break;
		}

		/* Find pushed return address behind the base pointer. */
		actual_eip = eip = READ_MEMORY(cpu, ebp + WORD_SIZE);
		eip = check_noreturn_function(cpu, eip);
		if (eip == 0) {
			break;
		}
		stack_ptr = ebp + (2 * WORD_SIZE);
		/* Suppress kernel frames if testing user, unless verbose enough. */
		if (splice_pre_vanish_trace(ls, st, eip)) {
			return st;
		} else if (!SUPPRESS_FRAME(eip)) {
			bool success = add_frame(st, eip, actual_eip);
			if (!success && wrong_cr3)
				return st;

			/* special-case termination condition -- _start */
			if (eip == GUEST_START) {
				break;
			}
		}

		if (rabbit > (unsigned int)(-4)) stop_ebp = rabbit;
		if (rabbit != stop_ebp) rabbit = READ_MEMORY(cpu, rabbit);
		if (rabbit == ebp) stop_ebp = ebp;
		if (rabbit > (unsigned int)(-4)) stop_ebp = rabbit;
		if (rabbit != stop_ebp) rabbit = READ_MEMORY(cpu, rabbit);
		if (rabbit == ebp) stop_ebp = ebp;
		ebp = READ_MEMORY(cpu, ebp);
		if (CHECK_JUNK_EBP_BELOW_TEXT(ebp)) {
			/* Some kernels allow terminal ebps to trail off into
			 * "junk values". Sometimes these are very small values.
			 * We can avoid emitting simics errors in these cases. */
			ebp = 0;
		} else if (KERNEL_MEMORY(stack_ptr) && USER_MEMORY(ebp) &&
			   // XXX: Hack to work around #194. Just hardcoding the
			   // most common values the crash was observed with.
			   ebp != 0x41414141 && ebp != 0xffff7000) {
			/* Base pointer chain crossed over into userspace,
			 * leaving stack pointer behind. This happens in pathos,
			 * whose syscall wrappers craft fake frames to help
			 * simics's tracer continue into userspace. */
#ifdef PATHOS_SYSCALL_IRET_DISTANCE
			/* Find iret frame. */
			stack_ptr += PATHOS_SYSCALL_IRET_DISTANCE;
			/* Not necessarily true. User thread could fault trying
			 * to call a kernel-space address. */
			//assert(USER_MEMORY(READ_MEMORY(cpu, stack_ptr)));
			assert(READ_MEMORY(cpu, stack_ptr + WORD_SIZE)
			       == SEGSEL_USER_CS);
			stack_ptr = READ_MEMORY(cpu, stack_ptr + (3 * WORD_SIZE));
			assert(USER_MEMORY(stack_ptr));
			// TODO: Decide what to do if pathos annotation missing?
			// Most likely, set some "skip_extra_frame" flag.
#endif
		}
	}

	return st;
}

/* As below but doesn't require duplicating the work of making a fresh stack
 * trace if you already have one. .*/
bool within_function_st(const struct stack_trace *st, unsigned int func,
			unsigned int func_end)
{
	const struct stack_frame *f;
	bool result = false;
	unsigned int i;
	ARRAY_LIST_FOREACH(&st->frames, i, f) {
		if (f->eip >= func && f->eip <= func_end) {
			result = true;
			break;
		}
	}
	return result;
}

/* Performs a stack trace to see if the current call stack has the given
 * function somewhere on it. */
bool within_function(struct ls_state *ls, unsigned int func, unsigned int func_end)
{
	/* Note while it may seem wasteful to malloc a bunch of times to make
	 * the stack trace, many (not all) of the mallocs are actually needed
	 * because the 'wrong cr3' end condition requires a symtable lookup. */
	struct stack_trace *st = stack_trace(ls);
	bool result = within_function_st(st, func, func_end);
	free_stack_trace(st);
	return result;
}

void dump_stack() {
	struct stack_trace *st = stack_trace(GET_LANDSLIDE());
	lsprintf(ALWAYS, "Stack trace: ");
	print_stack_trace(ALWAYS, st);
	printf(ALWAYS, "\n");
	free_stack_trace(st);
}
